( cd usr/lib ; rm -rf libgwrap-core-runtime.so.0 )
( cd usr/lib ; ln -sf libgwrap-core-runtime.so.0.0.1 libgwrap-core-runtime.so.0 )
( cd usr/lib ; rm -rf libgwrap-core-runtime.so )
( cd usr/lib ; ln -sf libgwrap-core-runtime.so.0.0.1 libgwrap-core-runtime.so )
( cd usr/lib ; rm -rf libgwrap-guile-runtime.so.0 )
( cd usr/lib ; ln -sf libgwrap-guile-runtime.so.0.0.2 libgwrap-guile-runtime.so.0 )
( cd usr/lib ; rm -rf libgwrap-guile-runtime.so )
( cd usr/lib ; ln -sf libgwrap-guile-runtime.so.0.0.2 libgwrap-guile-runtime.so )
( cd usr/lib ; rm -rf libgw-guile-standard.so.0 )
( cd usr/lib ; ln -sf libgw-guile-standard.so.0.0.0 libgw-guile-standard.so.0 )
( cd usr/lib ; rm -rf libgw-guile-standard.so )
( cd usr/lib ; ln -sf libgw-guile-standard.so.0.0.0 libgw-guile-standard.so )
( cd usr/lib ; rm -rf libgw-guile-gw-glib.so.0 )
( cd usr/lib ; ln -sf libgw-guile-gw-glib.so.0.0.0 libgw-guile-gw-glib.so.0 )
( cd usr/lib ; rm -rf libgw-guile-gw-glib.so )
( cd usr/lib ; ln -sf libgw-guile-gw-glib.so.0.0.0 libgw-guile-gw-glib.so )
