;; Generated by G-Wrap-TNG: an experimental Guile C API-wrapper engine.

(define-module (g-wrap gw standard)
  #:use-module (oop goops)
)

(dynamic-call "gw_init_wrapset_standard"
              (dynamic-link "libgw-guile-standard"))
(module-use! (module-public-interface (current-module)) (current-module))
